<?php

declare(strict_types=1);

namespace Drupal\localgov_waste_collection_whitespace_provider\Plugin\DataProvider;

use Drupal\localgov_waste_collection\DataProviderBase;
use GuzzleHttp\Client;

/**
 * Whitespace waste collection data provider.
 *
 * @DataProvider(
 *   id = "whitespace_data_provider",
 *   label = @Translation("Whitespace Data Provider")
 * )
 */
class WhitespaceDataProvider extends DataProviderBase {

  /**
   * Construct SOAP header XML.
   */
  private function getSoapHeader():string {
    $config = \Drupal::config('localgov_waste_collection_whitespace.settings');
    return '<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
                  xmlns:web="http://webservices.whitespacews.com/">
    <soapenv:Header>
        <wsse:Security soapenv:mustUnderstand="1"
                       xmlns:wsse="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"
                       xmlns:wsu="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">
            <wsse:UsernameToken>
                <wsse:Username>' . $config->get('username') . '</wsse:Username>
                <wsse:Password
                        Type="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordText">' . $config->get('password') . '</wsse:Password>
            </wsse:UsernameToken>
        </wsse:Security>
    </soapenv:Header>';
  }

  /**
   * Construct SOAP footer XML.
   */
  private function getSoapFooter():string {
    return '</soapenv:Envelope>';
  }

  /**
   * {@inheritdoc}
   */
  public function findAddressesByPostcode(string $postcode): array {
    $config = \Drupal::config('localgov_waste_collection_whitespace.settings');

    $properties = [];

    $soapBody = '
    <soapenv:Body>
        <web:GetAddresses>
            <web:getAddressInput>
                <web:Postcode> ' . $postcode . '</web:Postcode>
                <web:IncludeCommercialSites>0</web:IncludeCommercialSites>
            </web:getAddressInput>
        </web:GetAddresses>
    </soapenv:Body>
    ';

    $xmlRequest = $this->getSoapHeader() . $soapBody . $this->getSoapFooter();

    $client = new Client();

    $options = [
      'body'    => $xmlRequest,
      'headers' => [
        "SOAPAction" => "http://webservices.whitespacews.com/GetAddresses",
        "Content-Type" => "text/xml",
        "accept" => "*/*",
        "accept-encoding" => "gzip, deflate",
      ],
    ];

    $response = $client->request(
      'POST',
      $config->get('endpoint') . 'WSAPIService.svc/GetAddresses',
      $options
    );

    $xml = simplexml_load_string((string) $response->getBody());

    $addressesXml = $xml
      ->children('s', TRUE)
      ->Body
      ->children('http://webservices.whitespacews.com/')
      ->GetAddressesResponse
      ->GetAddressesResult
      ->Addresses;

    $json = json_encode($addressesXml);
    $addresses = json_decode($json, TRUE);

    if ($addresses) {
      if (!$addresses["Address"][0]) {
        $properties[$addresses["Address"]["AccountSiteUprn"]] = $this->formatAddress($addresses["Address"]);
      }
      else {
        foreach ($addresses["Address"] as $address) {
          $properties[$address["AccountSiteUprn"]] = $this->formatAddress($address);
        }
      }
      asort($properties, SORT_NUMERIC);
    }

    return $properties;
  }

  /**
   * {@inheritdoc}
   */
  public function getAddressFromUprn(string $uprn): ?string {
    $config = \Drupal::config('localgov_waste_collection_whitespace.settings');
    $address = NULL;

    $soapBody = '
    <soapenv:Body>
        <web:GetSiteInfo>
            <web:siteInfoInput>
                <web:Uprn> ' . $uprn . '</web:Uprn>
            </web:siteInfoInput>
        </web:GetSiteInfo>
    </soapenv:Body>
    ';

    $xmlRequest = $this->getSoapHeader() . $soapBody . $this->getSoapFooter();

    $client = \Drupal::service('http_client');

    $options = [
      'body'    => $xmlRequest,
      'headers' => [
        "SOAPAction" => "http://webservices.whitespacews.com/GetSiteInfo",
        "Content-Type" => "text/xml",
        "accept" => "*/*",
        "accept-encoding" => "gzip, deflate",
      ],
    ];

    $response = $client->request(
      'POST',
      $config->get('endpoint') . 'WSAPIService.svc/GetSiteInfo',
      $options
    );

    $xml = simplexml_load_string((string) $response->getBody());

    $addressXml = $xml
      ->children('s', TRUE)
      ->Body
      ->children('http://webservices.whitespacews.com/')
      ->GetSiteInfoResponse
      ->GetSiteInfoResult
      ->Site
      ->Site;

    $json = json_encode($addressXml);
    $address = $this->formatAddress(json_decode($json, TRUE));

    return $address;
  }

  /**
   * Format raw address.
   */
  private function formatAddress(array $address):string {
    $addressElements = [];
    $address = array_filter($address);
    if (isset($address["SiteName"])) {
      $addressElements[] = (string) $address["SiteName"];
    }
    if (isset($address["SiteAddressNumber"])) {
      $addressElements[] = str_replace('  ', ' ', $address["SiteAddressNumber"] . ' ' . ($address["SiteAddress1"] ?? '') . ' ' . $address["SiteAddress2"]);
    }
    if (isset($address["SiteTown"])) {
      $addressElements[] = (string) $address["SiteTown"];
    }
    $addressElements[] = (string) $address["SitePostCode"];

    return implode(', ', $addressElements);
  }

  /**
   * Get waste collections for a specified UPRN.
   *
   * @return array
   *   An array of waste collections.
   */
  public function getCollections(string $uprn): array {
    $config = \Drupal::config('localgov_waste_collection_whitespace.settings');

    // Temporary solution for API v11.0 - date range collections call not
    // available until v11.1, so we loop per month.
    // Call GetCollectionByUprnAndDate six times and merge results.
    $collections = [];
    $currentDate = new \DateTime();

    for ($monthOffset = 0; $monthOffset < 6; $monthOffset++) {
      $callDate = $currentDate->format('01/m/Y');
      $currentDate->modify('first day of next month');

      $soapBody = '
      <soapenv:Body>
        <web:GetCollectionByUprnAndDate>
          <web:getCollectionByUprnAndDateInput>
            <web:Uprn>' . $uprn . '</web:Uprn>
            <web:NextCollectionFromDate>' . $callDate . '</web:NextCollectionFromDate>
          </web:getCollectionByUprnAndDateInput>
        </web:GetCollectionByUprnAndDate>
      </soapenv:Body>
      ';

      $xmlRequest = $this->getSoapHeader() . $soapBody . $this->getSoapFooter();

      $client = \Drupal::service('http_client');

      $options = [
        'body' => $xmlRequest,
        'headers' => [
          "SOAPAction" => "http://webservices.whitespacews.com/GetCollectionByUprnAndDate",
          "Content-Type" => "text/xml",
          "accept" => "*/*",
          "accept-encoding" => "gzip, deflate",
        ],
      ];

      $response = $client->request(
        'POST',
        $config->get('endpoint') . 'WSAPIService.svc/GetCollectionByUprnAndDate',
        $options
      );

      $xml = simplexml_load_string((string) $response->getBody());
      $collectionsXml = $xml
        ->children('s', TRUE)
        ->Body
        ->children('http://webservices.whitespacews.com/')
        ->GetCollectionByUprnAndDateResponse
        ->GetCollectionByUprnAndDateResult
        ->Collections;

      $json = json_encode($collectionsXml);
      $monthCollections = json_decode($json, TRUE);

      foreach ($monthCollections["Collection"] as $collection) {
        $collections["Collection"][] = $collection;
      }
    }
    return $this->getCollectionsArray($collections);
  }

  /**
   * Construct an array of collections.
   */
  private function getCollectionsArray(array $collections): array {
    // There are many different bin types, this consolidates them so we can
    // display a summary.
    $binTypes = [
      'Residual Waste Collections' => ['label' => 'General waste', 'colour' => 'Grey'],
      'Dry Recycling Collection Service' => ['label' => 'Recyclable waste', 'colour' => 'Red'],
      'Garden Waste Collection Service' => ['label' => 'Garden waste', 'colour' => 'Brown/Green'],
    ];

    $collectionArray = [
      'weekly_collection' => [],
      'dates' => [],
    ];

    foreach ($collections["Collection"] as $collection) {
      // Replace date slashes with hyphens so date sorting will not assume US
      // format.
      $formattedDate = str_replace('/', '-', substr($collection["Date"], 0, 10));
      $holiday = NULL;
      $type = $binTypes[$collection["Service"]];
      $collectionArray["dates"][] = [
        "date" => $formattedDate,
        "bin" => $collection["Service"],
        "holiday" => $holiday,
        "type" => $type,
      ];
    }
    usort($collectionArray["dates"], static fn($a, $b) => strtotime($a["date"]) - strtotime($b["date"]));
    $collectionArray["dates"] = $this->removeDuplicateCollections($collectionArray["dates"]);
    return $collectionArray;
  }

  /**
   * Remove duplicates from the collection dates array.
   *
   * @param array $array
   *   Raw data from Whitespace.
   *
   * @return array
   *   De-duped data.
   */
  public function removeDuplicateCollections($array): array {
    $uniqueCollections = array_map("unserialize", array_unique(array_map("serialize", $array)));

    foreach ($uniqueCollections as $key => $value) {
      if (is_array($value)) {
        $uniqueCollections[$key] = self::removeDuplicateCollections($value);
      }
    }

    return $uniqueCollections;
  }

}
